#    Copyright 2021, 2022, 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from abc import ABCMeta, abstractmethod
from functools import cached_property
from collections import defaultdict
from typing import (
    Dict,
    Generic,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    Type,
    TypeVar,
)


from wsidicom.errors import WsiDicomMatchError, WsiDicomNotFoundError
from wsidicom.geometry import Size, SizeMm
from wsidicom.group import Group, Level
from wsidicom.instance import ImageType, WsiDataset, WsiInstance
from wsidicom.metadata.schema.dicom.wsi import WsiMetadataDicomSchema
from wsidicom.metadata.wsi import WsiMetadata
from wsidicom.uid import SlideUids

SeriesType = TypeVar("SeriesType", bound="Series")
GroupType = TypeVar("GroupType", Group, Level)


class Series(Generic[GroupType], metaclass=ABCMeta):
    """Represents a series of Groups with the same image flavor, e.g.
    pyramidal levels, labels, or overviews.
    """

    def __init__(self, groups: Iterable[GroupType]):
        """Create a Series from list of Groups.

        Parameters
        ----------
        groups: Iterable[Group]
            List of groups to include in the series.
        """
        self._groups = list(groups)
        if len(self._groups) != 0 and self._groups[0].uids is not None:
            self._uids = self._validate_series(self._groups)
        else:
            self._uids = None

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self._groups})"

    def __str__(self) -> str:
        return f"{type(self).__name__} of groups {self._groups}"

    def __getitem__(self, index: int) -> GroupType:
        return self._groups[index]

    def get(self, index: int) -> GroupType:
        """Get group by index.

        Parameters
        ----------
        index: int
            Index in series to get

        Returns
        -------
        Group
            The group at index in the series
        """
        try:
            return self[index]
        except IndexError:
            raise WsiDicomNotFoundError(f"Group index {index}", "series")

    def __len__(self) -> int:
        return len(self._groups)

    @property
    @abstractmethod
    def image_type(self) -> ImageType:
        """Should return the wsi type of the series ('VOLUME', 'LABEL', or
        'OVERVIEW'"""
        raise NotImplementedError()

    @property
    def groups(self) -> List[GroupType]:
        """Return contained groups."""
        return self._groups

    @property
    def uids(self) -> Optional[SlideUids]:
        """Return uids."""
        return self._uids

    @property
    def mpps(self) -> List[SizeMm]:
        """Return contained mpp (um/px)."""
        return [group.mpp for group in self if group.mpp is not None]

    @property
    def datasets(self) -> List[WsiDataset]:
        """Return contained datasets."""

        series_datasets = [series.datasets for series in self]
        return [dataset for sublist in series_datasets for dataset in sublist]

    @property
    def instances(self) -> List[WsiInstance]:
        """Return contained instances"""
        series_instances = [series.instances.values() for series in self]
        return [instance for sublist in series_instances for instance in sublist]

    @cached_property
    def metadata(self) -> WsiMetadata:
        return WsiMetadataDicomSchema().load(self.datasets[0])

    @classmethod
    def open(
        cls: Type[SeriesType], instances: Iterable[WsiInstance]
    ) -> Optional[SeriesType]:
        """Return series created from instances.

        Parameters
        ----------
        instances: Iterable[WsiInstance]
            Instances to create series from.

        Returns
        -------
        Optional[SeriesType]
            Created series.
        """
        instances_grouped_by_size = cls._group_instances_by_size(instances)
        groups = [Group(instances) for instances in instances_grouped_by_size]
        if len(groups) == 0:
            return None
        return cls(groups)

    @classmethod
    def _group_instances_by_size(
        cls, instances: Iterable[WsiInstance]
    ) -> Iterator[List[WsiInstance]]:
        """Return instances grouped and sorted by image size.

        Parameters
        ----------
        instances: Iterable[WsiInstance]
            Instances to group by image size.

        Returns
        -------
        Iterator[List[WsiInstance]]:
            Instances grouped by size.

        """
        grouped_instances: Dict[Size, List[WsiInstance]] = defaultdict(list)
        for instance in instances:
            grouped_instances[instance.size].append(instance)
        return (
            grouped_instances[key]
            for key in sorted(
                grouped_instances.keys(), key=lambda size: size.width, reverse=True
            )
        )

    def _validate_series(self, groups: Sequence[GroupType]) -> Optional[SlideUids]:
        """Check that no files or instances in series is duplicate and that
        all groups in series matches.
        Raises WsiDicomMatchError otherwise.
        Returns the matching base uid. If list of groups is empty, return None.

        Parameters
        ----------
        groups: Union[Sequence[Group], Sequence[Level]]
            List of groups or levels to check

        Returns
        -------
        Optional[SlideUids]:
            Matching uids
        """
        WsiDataset.check_duplicate_dataset(self.datasets, self)
        WsiInstance.check_duplicate_instance(self.instances, self)

        try:
            base_group = groups[0]
            if base_group.image_type != self.image_type:
                raise WsiDicomMatchError(str(base_group), str(self))
            for group in groups[1:]:
                if not group.matches(base_group):
                    raise WsiDicomMatchError(str(group), str(self))
            return base_group.uids
        except IndexError:
            return None


class Overviews(Series[Group]):
    """Represents a series of Groups of the overview wsi flavor."""

    @property
    def image_type(self) -> ImageType:
        return ImageType.OVERVIEW


class Labels(Series[Group]):
    """Represents a series of Groups of the label wsi flavor."""

    @property
    def image_type(self) -> ImageType:
        return ImageType.LABEL



#    Copyright 2021, 2022, 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Module implementing generic ImageData for accessing image data."""

from abc import ABCMeta, abstractmethod
from typing import (
    Callable,
    Iterable,
    Iterator,
    List,
    Optional,
    Tuple,
    Union,
)

from PIL import Image as Pillow
from PIL.Image import Image
from pydicom.uid import UID

from wsidicom.codec import Encoder
from wsidicom.config import settings
from wsidicom.errors import WsiDicomOutOfBoundsError
from wsidicom.geometry import Point, Region, Size, SizeMm
from wsidicom.metadata.image import ImageCoordinateSystem
from wsidicom.thread import ConditionalThreadPoolExecutor


class ImageData(metaclass=ABCMeta):
    """
    Generic class for image data.

    Can be inherited to implement support for other image/file formats. Subclasses
    should implement properties to get transfer_syntax, image_size, tile_size,
    pixel_spacing,  samples_per_pixel, and photometric_interpretation and methods
    _get_decoded_tile(), _get_encoded_tile(), and close(). Optionally the methods
    _get_decoded_tiles() and _get_encoded_tiles() can be overridden for more efficient
    fetching of multiple tiles.

    Additionally properties focal_planes and/or optical_paths should be
    overridden if multiple focal planes or optical paths are implemented.
    """

    _default_z: Optional[float] = None
    _blank_tile: Optional[Image] = None
    _encoded_blank_tile: Optional[bytes] = None

    def __init__(self, encoder: Encoder):
        self._encoder = encoder

    @property
    @abstractmethod
    def transfer_syntax(self) -> UID:
        """Return the uid of the transfer syntax of the image."""
        raise NotImplementedError()

    @property
    @abstractmethod
    def image_size(self) -> Size:
        """Return the pixel size of the image."""
        raise NotImplementedError()

    @property
    @abstractmethod
    def tile_size(self) -> Size:
        """Return the pixel size of a tile in the image, or image size if not tiled."""
        raise NotImplementedError()

    @property
    @abstractmethod
    def pixel_spacing(self) -> Optional[SizeMm]:
        """Return the size of the pixels in mm/pixel."""
        raise NotImplementedError()

    @property
    @abstractmethod
    def samples_per_pixel(self) -> int:
        """Return number of samples per pixel (e.g. 3 for RGB)."""
        raise NotImplementedError()

    @property
    @abstractmethod
    def photometric_interpretation(self) -> str:
        """Return the photophotometric interpretation of the image data."""
        raise NotImplementedError()

    @property
    @abstractmethod
    def bits(self) -> int:
        """Should return the number of bits stored for each sample."""
        raise NotImplementedError()

    @property
    @abstractmethod
    def lossy_compressed(self) -> bool:
        """Should return True if the image has been lossy compressed."""
        raise NotImplementedError()

    @property
    @abstractmethod
    def image_coordinate_system(self) -> Optional[ImageCoordinateSystem]:
        """Should return the image origin of the image data."""
        raise NotImplementedError()

    @abstractmethod
    def _get_decoded_tile(self, tile_point: Point, z: float, path: str) -> Image:
        """
        Return Pillow image for tile.

        Parameters
        ----------
        tile: Point
            Tiles to get.
        z: float
            Z coordinate.
        path: str
            Optical path.

        Returns
        -------
        Image
            Tile as Pillow Image.
        """
        raise NotImplementedError()

    @abstractmethod
    def _get_encoded_tile(self, tile: Point, z: float, path: str) -> bytes:
        """
        Return bytes for tile.

        Parameters
        ----------
        tile: Point
            Tiles to get.
        z: float
            Z coordinate.
        path: str
            Optical path.

        Returns
        -------
        bytes
            Tile as bytes.
        """
        raise NotImplementedError()

    @property
    def tiled_size(self) -> Size:
        """
        The size (columns and rows of tiles) of the image when divided into tiles.

        Equals (1, 1) if image is not tiled.
        """
        return self.image_size.ceil_div(self.tile_size)

    @property
    def image_region(self) -> Region:
        """The pixel region the image covers."""
        return Region(Point(0, 0), self.image_size)

    @property
    def focal_planes(self) -> List[float]:
        """Focal planes available in the image defined in um."""
        return [0.0]

    @property
    def optical_paths(self) -> List[str]:
        """Optical paths available in the image."""
        return ["0"]

    @property
    def image_mode(self) -> str:
        """Return Pillow image mode (e.g. RGB) for image data."""
        if self.samples_per_pixel == 1:
            if self.bits == 8:
                return "L"
            elif self.bits == 16:
                return "I"
        elif self.samples_per_pixel == 3:
            return "RGB"
        raise NotImplementedError()

    @property
    def blank_color(self) -> Union[int, Tuple[int, int, int]]:
        """Return grey level or RGB background color."""
        return self._get_blank_color(self.photometric_interpretation)

    def pretty_str(self, indent: int = 0, depth: Optional[int] = None) -> str:
        """Return pretty string of object."""
        return str(self)

    @property
    def default_z(self) -> float:
        """
        The focal plane to use as default.

        Return single defined focal plane (in um) if only one focal plane defined.
        Return the middle focal plane if several focal planes are defined.
        """
        if self._default_z is None:
            default = 0
            if len(self.focal_planes) > 1:
                smallest = min(self.focal_planes)
                largest = max(self.focal_planes)
                middle = (largest - smallest) / 2
                default = min(
                    range(len(self.focal_planes)),
                    key=lambda i: abs(self.focal_planes[i] - middle),
                )

            self._default_z = self.focal_planes[default]

        return self._default_z

    @property
    def default_path(self) -> str:
        """Return the first defined optical path as default optical path identifier."""
        return self.optical_paths[0]

    @property
    def plane_region(self) -> Region:
        return Region(position=Point(0, 0), size=self.tiled_size)

    @property
    def blank_tile(self) -> Image:
        """Return background tile."""
        if self._blank_tile is None:
            self._blank_tile = self._create_blank_tile()
        return self._blank_tile

    @property
    def blank_encoded_tile(self) -> bytes:
        """Return encoded background tile."""
        if self._encoded_blank_tile is None:
            self._encoded_blank_tile = self.encoder.encode(self.blank_tile)
        return self._encoded_blank_tile

    @property
    def encoder(self) -> Encoder:
        return self._encoder

    def get_decoded_tiles(
        self, tiles: Iterable[Point], z: float, path: str
    ) -> Iterator[Image]:
        """
        Return Pillow images for tiles.

        Parameters
        ----------
        tiles: Iterable[Point]
            Tiles to get.
        z: float
            Z coordinate.
        path: str
            Optical path.

        Returns
        -------
        Iterator[Image]
            Tiles as Images.
        """
        return self._get_decoded_tiles(tiles, z, path)

    def get_encoded_tiles(
        self,
        tiles: Iterable[Point],
        z: float,
        path: str,
        scale: int = 1,
        reencoder: Optional[Encoder] = None,
    ) -> Iterator[bytes]:
        """
        Return bytes for tiles.

        Parameters
        ----------
        tiles: Iterable[Point]
            Tiles to get.
        z: float
            Z coordinate.
        path: str
            Optical path.
        scale: int = 1
            Scale to use for downscaling.
        reencoder: Optional[Encoder] = None
            Encoder to use for re-encoding. If None do not re-encode tiles.

        Returns
        -------
        Iterator[Image]
            Tiles as Images.
        """
        if reencoder is not None:
            return self._get_reencoded_tiles(tiles, z, path, scale, reencoder)
        if scale == 1:
            return self._get_encoded_tiles(tiles, z, path)
        return (self.get_scaled_encoded_tile(tile, z, path, scale) for tile in tiles)

    def get_scaled_tile(
        self,
        scaled_tile_point: Point,
        z: float,
        path: str,
        scale: int,
        workers: int = 1,
    ) -> Image:
        """
        Return scaled tile as Pillow image.

        Parameters
        ----------
        scaled_tile_point: Point,
            Scaled position of tile to get.
        z: float
            Z coordinate.
        path: str
            Optical path.
        scale: int
            Scale to use for downscaling.
        workers: int = 1
            Workers to use for threading. Default to not use threading as method is
            likely already used in a threading context.

        Returns
        -------
        Image
            Scaled tiled as Pillow image.
        """
        image = Pillow.new(
            mode=self.image_mode,
            size=(self.tile_size * scale).to_tuple(),
            color=self.blank_color,
        )
        # Get decoded tiles for the region covering the scaled tile
        # in the image data
        tile_region = Region(scaled_tile_point * scale, Size(1, 1) * scale)
        tile_region = tile_region.crop(self.tiled_size)

        def paste(image: Image, tile_point: Point, tile: Image):
            image_coordinate = (tile_point - tile_region.start) * self.tile_size
            image.paste(tile, image_coordinate.to_tuple())

        self._paste_tiles(
            image,
            tile_region,
            z,
            path,
            paste,
            workers,
        )

        return image.resize(
            self.tile_size.to_tuple(), resample=settings.pillow_resampling_filter
        )

    def get_tile(self, tile_point: Point, z: float, path: str) -> Image:
        """
        Get tile as Pillow image.

        If frame is inside tile geometry but no tile exists in frame data (sparse)
        returns blank image. Crops tile to be inside image boundary.

        Parameters
        ----------
        tile_point: Point
            Tile x, y coordinate.
        z: float
            Z coordinate.
        path: str
            Optical path.


        Returns
        -------
        Image
            Tile image.
        """
        tile = self._get_decoded_tile(tile_point, z, path)
        return self._crop_tile(tile_point, tile)

    def get_tiles(self, tiles: Iterable[Point], z: float, path: str) -> Iterator[Image]:
        return (
            self._crop_tile(tile_point, tile)
            for tile_point, tile in zip(tiles, self.get_decoded_tiles(tiles, z, path))
        )

    def get_scaled_encoded_tile(
        self, scaled_tile_point: Point, z: float, path: str, scale: int
    ) -> bytes:
        """
        Return scaled tile as bytes.

        Parameters
        ----------
        scaled_tile_point: Point,
            Scaled position of tile to get.
        z: float
            Z coordinate.
        path: str
            Optical path.
        Scale: int
            Scale to use for downscaling.

        Returns
        -------
        bytes
            Scaled tile as bytes.
        """
        image = self.get_scaled_tile(scaled_tile_point, z, path, scale)
        return self.encoder.encode(image)

    def get_encoded_tile(self, tile: Point, z: float, path: str) -> bytes:
        """
        Get tile as bytes.

        If frame is inside tile geometry but no tile exists in frame data (sparse)
        returns encoded blank image. Crops tile to be inside image boundary.

        Parameters
        ----------
        tile: Point
            Tile x, y coordinate.
        z: float
            Z coordinate.
        path: str
            Optical path.

        Returns
        -------
        bytes
            Tile image as bytes.
        """
        # Check if tile is an edge tile that should be cropped
        cropped_tile_region = self.image_region.inside_crop(tile, self.tile_size)
        if cropped_tile_region.size == self.tile_size:
            return self._get_encoded_tile(tile, z, path)
        image = self._get_decoded_tile(tile, z, path)
        image.crop(box=cropped_tile_region.box_from_origin)
        return self.encoder.encode(image)

    def stitch_tiles(self, region: Region, path: str, z: float, threads: int) -> Image:
        """Stitches tiles together to form requested image.

        Parameters
        ----------
        region: Region
             Pixel region to stitch to image
        path: str
            Optical path
        z: float
            Z coordinate
        threads: int
            Number of threads to use for read.

        Returns
        -------
        Image
            Stitched image
        """

        tile_region = self._get_tile_range(region, z, path)
        if tile_region.size.area == 1:
            tile = self._get_decoded_tile(tile_region.start, z, path)
            tile_crop = region.inside_crop(tile_region.start, self.tile_size)
            if tile_crop.size != self.tile_size:
                tile = tile.crop(box=tile_crop.box)
            return tile
        image = Pillow.new(mode=self.image_mode, size=region.size.to_tuple())
        # The tiles are cropped prior to pasting. This offset is the equal to the first
        # (upper left) tiles size, and is added to the image coordinate for tiles not
        # in the first row or column.
        offset = (tile_region.start * self.tile_size) - region.start

        def paste(image: Image, tile_point: Point, tile: Image):
            image_coordinate = Point(
                offset.x * (tile_point.x != tile_region.start.x),
                offset.y * (tile_point.y != tile_region.start.y),
            )
            image_coordinate += (tile_point - tile_region.start) * self.tile_size
            tile_crop = region.inside_crop(tile_point, self.tile_size)
            if tile_crop.size != self.tile_size:
                tile = tile.crop(box=tile_crop.box)
            image.paste(tile, image_coordinate.to_tuple())

        self._paste_tiles(
            image,
            tile_region,
            z,
            path,
            paste,
            threads,
        )
        return image

    def valid_tiles(self, region: Region, z: float, path: str) -> bool:
        """
        Check if tile region is inside image and z coordinate and optical path exists.

        Parameters
        ----------
        region: Region
            Tile region.
        z: float
            Z coordinate.
        path: str
            Optical path.

        Returns
        -------
        bool
            True if tile region is inside image and z coordinate and optical path
            exists.
        """
        return (
            region.is_inside(self.plane_region)
            and (z in self.focal_planes)
            and (path in self.optical_paths)
        )

    def _get_decoded_tiles(
        self, tiles: Iterable[Point], z: float, path: str
    ) -> Iterator[Image]:
        """
        Return Pillow images for tiles. Implementations can override this with a more
        efficient method for getting multiple tiles.

        Parameters
        ----------
        tiles: Iterable[Point]
            Tiles to get.
        z: float
            Z coordinate.
        path: str
            Optical path.

        Returns
        -------
        Iterator[Image]
            Iterator of tiles as Images.
        """
        return (self._get_decoded_tile(tile, z, path) for tile in tiles)

    def _get_encoded_tiles(
        self, tiles: Iterable[Point], z: float, path: str
    ) -> Iterator[bytes]:
        """
        Return bytes for tiles. Implementations can override this with a more efficient
        method for getting multiple tiles.

        Parameters
        ----------
        tiles: Iterable[Point]
            Tiles to get.
        z: float
            Z coordinate.
        path: str
            Optical path.

        Returns
        -------
        Iterator[bytes]
            Iterator of tiles as bytes.
        """
        return (self._get_encoded_tile(tile, z, path) for tile in tiles)

    def _crop_tile(self, tile_point: Point, tile: Image) -> Image:
        tile_crop = self.image_region.inside_crop(tile_point, self.tile_size)
        # Check if tile is an edge tile that should be cropped
        if tile_crop.size != self.tile_size:
            return tile.crop(box=tile_crop.box)
        return tile

    def _paste_tiles(
        self,
        image: Image,
        tile_region: Region,
        z: float,
        path: str,
        paste_method: Callable[[Image, Point, Image], None],
        threads: int,
    ):
        """
        Paste tiles in region using method.

        Use threading if number of tiles to paste
        is larger than one and requested worker count is more than one.

        Parameters
        ----------
        image: Image
            Image to paste into.
        tile_region: Region
            Tile region of tiles to paste.
        z: float
            Z coordinate.
        path: str
            Optical path.
        paste_method: Callable[[Image, Point, Image], None]
            Method that accepts a image, a tile point and tile to paste and returns
            None.
        threads: int
            Number of workers to use.
        """

        def thread_paste(tile_points: Iterable[Point]) -> None:
            tile_points = list(tile_points)
            for tile_point, tile in zip(
                tile_points, self.get_decoded_tiles(tile_points, z, path)
            ):
                paste_method(image, tile_point, tile)

        with ConditionalThreadPoolExecutor(
            max_workers=threads, force_iteration=True
        ) as pool:
            pool.map(thread_paste, tile_region.chunked_iterate_all(threads))

    def _get_reencoded_tiles(
        self,
        tiles: Iterable[Point],
        z: float,
        path: str,
        scale: int,
        reencoder: Encoder,
    ) -> Iterator[bytes]:
        """
        Return re-encoded bytes for tiles.

        Parameters
        ----------
        tiles: Iterable[Point]
            Tiles to get.
        z: float
            Z coordinate.
        path: str
            Optical path.
        scale: int
            Scale to use for downscaling.
        reencoder: Encoder
            Encoder to use for re-encoding.

        """
        if scale == 1:
            decoded_tiles = self._get_decoded_tiles(tiles, z, path)
        else:
            decoded_tiles = (
                self.get_scaled_tile(tile, z, path, scale) for tile in tiles
            )
        return (reencoder.encode(tile) for tile in decoded_tiles)

    @staticmethod
    def _get_blank_color(
        photometric_interpretation: str,
    ) -> Union[int, Tuple[int, int, int]]:
        """Return color to use blank tiles.

        Parameters
        ----------
        photometric_interpretation: str
            The photomoetric interpretation of the dataset

        Returns
        -------
        Union[int, Tuple[int, int, int]]
            Grey level or RGB color,

        """
        BLACK = 0
        WHITE = 255
        if photometric_interpretation == "MONOCHROME2":
            return BLACK
        return (WHITE, WHITE, WHITE)

    def _create_blank_tile(self) -> Image:
        """Create blank tile for instance.

        Returns
        -------
        Image
            Blank tile image
        """
        return Pillow.new(
            mode=self.image_mode,
            size=self.tile_size.to_tuple(),
            color=self.blank_color,
        )

    def _get_tile_range(self, pixel_region: Region, z: float, path: str) -> Region:
        """Return range of tiles to cover pixel region.

        Parameters
        ----------
        pixel_region: Region
            Pixel region of tiles to get
        z: float
            Z coordinate of tiles to get
        path: str
            Optical path identifier of tiles to get

        Returns
        -------
        Region
            Region of tiles for stitching image
        """

        tile_region = Region.from_points(
            pixel_region.start // self.tile_size,
            (pixel_region.end - 1) // self.tile_size + 1,
        )
        if not self.valid_tiles(tile_region, z, path):
            raise WsiDicomOutOfBoundsError(
                f"Tile region {tile_region}", f"tiled size {self.tiled_size}"
            )
        return tile_region



#    Copyright 2021, 2022, 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from functools import lru_cache
from typing import List, OrderedDict, Sequence, Union

from pydicom.uid import UID
from upath import UPath

from wsidicom.codec import Codec
from wsidicom.errors import WsiDicomNotFoundError
from wsidicom.file.io import WsiDicomReader
from wsidicom.instance import WsiDicomImageData


class WsiDicomFileImageData(WsiDicomImageData):
    """
    Represents image data read from dicom file(s).

    Image data can be sparsly or fully tiled and/or concatenated.
    """

    def __init__(
        self, readers: Union[WsiDicomReader, Sequence[WsiDicomReader]]
    ) -> None:
        """
        Create WsiDicomFileImageData from frame data from readers.

        Parameters
        ----------
        readers: Union[WsiDicomReader, Sequence[WsiDicomReader]]
            Single or list of WsiDicomReader containing frame data.
        """
        if not isinstance(readers, Sequence):
            readers = [readers]

        # Key is frame offset
        self._readers = OrderedDict(
            (file.frame_offset, file)
            for file in sorted(readers, key=lambda file: file.frame_offset)
        )
        self._transfer_syntax = readers[0].transfer_syntax
        dataset = readers[0].dataset
        codec = Codec.create(
            self.transfer_syntax,
            dataset.samples_per_pixel,
            dataset.bits,
            dataset.tile_size,
            dataset.photometric_interpretation,
        )
        super().__init__([file.dataset for file in self._readers.values()], codec)

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self._readers.values()})"

    def __str__(self) -> str:
        return f"{type(self).__name__} of files {self._readers.values()}"

    @property
    def files(self) -> List[UPath]:
        return [
            reader.filepath
            for reader in self._readers.values()
            if reader.filepath is not None
        ]

    @property
    def transfer_syntax(self) -> UID:
        """The uid of the transfer syntax of the image."""
        return self._transfer_syntax

    @lru_cache
    def _get_reader(self, frame_index: int) -> WsiDicomReader:
        """
        Return file containing frame index.

        Raises WsiDicomNotFoundError if frame is not found.

        Parameters
        ----------
        frame_index: int
             Frame index to get

        Returns
        -------
        WsiDicomFile
            File containing the frame
        """
        for frame_offset, file in self._readers.items():
            if (
                frame_index < frame_offset + file.frame_count
                and frame_index >= frame_offset
            ):
                return file

        raise WsiDicomNotFoundError(f"Frame index {frame_index}", "instance")

    def _get_tile_frame(self, frame_index: int) -> bytes:
        """Return tile frame for frame index.

        Parameters
        ----------
        frame_index: int
             Frame index to get

        Returns
        -------
        bytes
            The frame in bytes
        """
        reader = self._get_reader(frame_index)
        tile_frame = reader.read_frame(frame_index)
        return tile_frame



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Json schema for Optical path model."""

import dataclasses
from typing import Any, Dict, Iterable, Mapping, Type, Union

from marshmallow import Schema, fields, post_load

from wsidicom.conceptcode import (
    IlluminationCode,
    IlluminationColorCode,
    ImagePathFilterCode,
    LenseCode,
    LightPathFilterCode,
)
from wsidicom.metadata.optical_path import (
    ConstantLutSegment,
    DiscreteLutSegment,
    ImagePathFilter,
    LightPathFilter,
    LinearLutSegment,
    Lut,
    LutSegment,
    Objectives,
    OpticalPath,
)
from wsidicom.metadata.schema.common import LoadingSchema
from wsidicom.metadata.schema.json.fields import (
    FileLoadingField,
    JsonFieldFactory,
    NpUIntDTypeField,
)


class BaseLutSegmentJsonSchema(Schema):
    _load_class: Type[Union[LinearLutSegment, ConstantLutSegment, DiscreteLutSegment]]

    @post_load
    def post_load(self, data: Dict[str, Any], **kwargs) -> LutSegment:
        """Return a object of given load class using the defined dataclass fields."""
        return self._load_class(
            **{
                field.name: data[field.name]
                for field in dataclasses.fields(self._load_class)
                if field.name in data
            }
        )


class DiscreteLutSegmentJsonSchema(BaseLutSegmentJsonSchema):
    values = fields.List(fields.Integer())
    _load_class = DiscreteLutSegment


class LinearLutSegmentJsonSchema(BaseLutSegmentJsonSchema):
    start_value = fields.Integer()
    end_value = fields.Integer()
    length = fields.Integer()
    _load_class = LinearLutSegment


class ConstantLutSegmentJsonSchema(BaseLutSegmentJsonSchema):
    value = fields.Integer()
    length = fields.Integer()
    _load_class = ConstantLutSegment


class LutSegmentJsonSchema(Schema):
    """Mapping segment type to schema."""

    _type_to_schema_mapping: Dict[Type[LutSegment], Type[Schema]] = {
        DiscreteLutSegment: DiscreteLutSegmentJsonSchema,
        LinearLutSegment: LinearLutSegmentJsonSchema,
        ConstantLutSegment: ConstantLutSegmentJsonSchema,
    }

    """Mapping key in serialized segment to schema."""
    _key_to_schema_mapping: Dict[str, Type[Schema]] = {
        "values": DiscreteLutSegmentJsonSchema,
        "start_value": LinearLutSegmentJsonSchema,
        "value": ConstantLutSegmentJsonSchema,
    }

    def dump(
        self,
        data: Union[LutSegment, Iterable[LutSegment]],
        **kwargs,
    ):
        if isinstance(data, LutSegment):
            return self._subschema_dump(data)
        return [self._subschema_dump(item) for item in data]

    def load(
        self,
        data: Union[Mapping[str, Any], Iterable[Mapping[str, Any]]],
        **kwargs,
    ):
        if isinstance(data, Mapping):
            return self._subschema_load(data)
        return [self._subschema_load(step) for step in data]

    def _subschema_load(self, segment: Mapping) -> LutSegment:
        """Select a schema and load and return step using the schema."""
        try:
            schema = next(
                schema
                for key, schema in self._key_to_schema_mapping.items()
                if key in segment
            )
        except StopIteration:
            raise NotImplementedError()
        loaded = schema().load(segment, many=False)
        assert isinstance(loaded, LutSegment)
        return loaded

    def _subschema_dump(self, segment: LutSegment):
        """Select a schema and dump the step using the schema."""
        schema = self._type_to_schema_mapping[type(segment)]
        return schema().dump(segment, many=False)


class LutJsonSchema(LoadingSchema[Lut]):
    red = fields.List(fields.Nested(LutSegmentJsonSchema()))
    green = fields.List(fields.Nested(LutSegmentJsonSchema()))
    blue = fields.List(fields.Nested(LutSegmentJsonSchema()))
    data_type = NpUIntDTypeField(data_key="bits")

    @property
    def load_type(self):
        return Lut


class LightPathFilterJsonSchema(LoadingSchema[LightPathFilter]):
    filters = fields.List(JsonFieldFactory.concept_code(LightPathFilterCode)())
    nominal = fields.Float(allow_none=True)
    low_pass = fields.Float(allow_none=True)
    high_pass = fields.Float(allow_none=True)

    @property
    def load_type(self):
        return LightPathFilter


class ImagePathFilterJsonSchema(LoadingSchema[ImagePathFilter]):
    filters = fields.List(JsonFieldFactory.concept_code(ImagePathFilterCode)())
    nominal = fields.Float(allow_none=True)
    low_pass = fields.Float(allow_none=True)
    high_pass = fields.Float(allow_none=True)

    @property
    def load_type(self):
        return ImagePathFilter


class ObjectivesJsonSchema(LoadingSchema[Objectives]):
    """Set of lens conditions for optical path"""

    lenses = fields.List(JsonFieldFactory.concept_code(LenseCode), allow_none=True)
    condenser_power = fields.Float(allow_none=True)
    objective_power = fields.Float(allow_none=True)
    objective_numerical_aperture = fields.Float(allow_none=True)

    @property
    def load_type(self):
        return Objectives


class OpticalPathJsonSchema(Schema):
    """Optical path. Icc profile is not included but can be loaded from context."""

    identifier = fields.String(allow_none=True)
    description = fields.String(allow_none=True)
    illumination_types = fields.List(
        JsonFieldFactory.concept_code(IlluminationCode)(), allow_none=True
    )
    illumination = JsonFieldFactory.float_or_concept_code(IlluminationColorCode)(
        allow_none=True
    )
    lut = fields.Nested(LutJsonSchema(), allow_none=True)
    icc_profile = FileLoadingField(load_only=True, load_default=None)
    light_path_filter = fields.Nested(LightPathFilterJsonSchema(), allow_none=True)
    image_path_filter = fields.Nested(ImagePathFilterJsonSchema(), allow_none=True)
    objective = fields.Nested(ObjectivesJsonSchema(), allow_none=True)

    @post_load
    def load_to_object(self, data: Dict[str, Any], **kwargs):
        icc_profile = self.context.get("icc_profile", None)
        if icc_profile is not None:
            if isinstance(icc_profile, bytes):
                data["icc_profile"] = icc_profile
            elif isinstance(icc_profile, dict):
                identifier = data.get("identifier", None)
                if identifier in icc_profile:
                    data["icc_profile"] = icc_profile[identifier]
        return OpticalPath(**data)



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Json schema for complete WsiMetadata."""

from marshmallow import fields

from wsidicom.metadata.equipment import Equipment
from wsidicom.metadata.image import Image
from wsidicom.metadata.label import Label
from wsidicom.metadata.patient import Patient
from wsidicom.metadata.schema.common import LoadingSchema
from wsidicom.metadata.schema.json.equipment import EquipmentJsonSchema
from wsidicom.metadata.schema.json.fields import UidJsonField
from wsidicom.metadata.schema.json.image import ImageJsonSchema
from wsidicom.metadata.schema.json.label import LabelJsonSchema
from wsidicom.metadata.schema.json.optical_path import OpticalPathJsonSchema
from wsidicom.metadata.schema.json.patient import PatientJsonSchema
from wsidicom.metadata.schema.json.series import SeriesJsonSchema
from wsidicom.metadata.schema.json.slide import SlideJsonSchema
from wsidicom.metadata.schema.json.study import StudyJsonSchema
from wsidicom.metadata.series import Series
from wsidicom.metadata.slide import Slide
from wsidicom.metadata.study import Study
from wsidicom.metadata.wsi import WsiMetadata


class WsiMetadataJsonSchema(LoadingSchema[WsiMetadata]):
    study = fields.Nested(StudyJsonSchema(), load_default=Study())
    series = fields.Nested(SeriesJsonSchema(), load_default=Series())
    patient = fields.Nested(PatientJsonSchema(), load_default=Patient())
    equipment = fields.Nested(EquipmentJsonSchema(), load_default=Equipment())
    optical_paths = fields.List(fields.Nested(OpticalPathJsonSchema()), load_default=[])
    slide = fields.Nested(SlideJsonSchema(), load_default=Slide())
    label = fields.Nested(LabelJsonSchema(), load_default=Label())
    image = fields.Nested(ImageJsonSchema(), load_default=Image())
    frame_of_reference_uid = UidJsonField(allow_none=True)
    dimension_organization_uids = fields.List(UidJsonField(), allow_none=True)

    @property
    def load_type(self):
        return WsiMetadata



#    Copyrigh 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from enum import Enum


class OffsetTableType(Enum):
    """Offset table type."""

    NONE = "none"
    EMPTY = "empty"
    BASIC = "BOT"
    EXTENDED = "EOT"
    TIFF = "TIFF"

    @classmethod
    def from_string(cls, offset_table: str) -> "OffsetTableType":
        """Return OffsetTableType parsed from string."""
        if offset_table.strip().lower() == "none":
            return OffsetTableType.NONE
        if offset_table.strip().lower() == "empty":
            return OffsetTableType.EMPTY
        if offset_table.strip().lower() == "eot":
            return OffsetTableType.EXTENDED
        if offset_table.strip().lower() == "bot":
            return OffsetTableType.BASIC
        if offset_table.strip().lower() == "tiff":
            return OffsetTableType.TIFF
        raise ValueError(f"Unknown offset table type: {offset_table}")



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""DICOM schema for Equipment model."""

from typing import Type

from pydicom.valuerep import VR

from wsidicom.metadata.equipment import Equipment
from wsidicom.metadata.schema.dicom.defaults import Defaults
from wsidicom.metadata.schema.dicom.fields import (
    DefaultingDicomField,
    ListDicomField,
    StringDicomField,
)
from wsidicom.metadata.schema.dicom.schema import ModuleDicomSchema


class EquipmentDicomSchema(ModuleDicomSchema[Equipment]):
    manufacturer = DefaultingDicomField(
        StringDicomField(VR.LO),
        dump_default=Defaults.string,
        load_default=None,
        data_key="Manufacturer",
    )
    model_name = DefaultingDicomField(
        StringDicomField(VR.LO),
        dump_default=Defaults.string,
        load_default=None,
        data_key="ManufacturerModelName",
    )
    device_serial_number = DefaultingDicomField(
        StringDicomField(VR.LO),
        dump_default=Defaults.string,
        load_default=None,
        data_key="DeviceSerialNumber",
    )
    software_versions = DefaultingDicomField(
        ListDicomField(StringDicomField(VR.LO)),
        dump_default=[Defaults.string],
        load_default=None,
        data_key="SoftwareVersions",
    )

    @property
    def load_type(self) -> Type[Equipment]:
        return Equipment

    @property
    def module_name(self) -> str:
        return "equipment"



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""DICOM schema for Study model."""

from typing import Type

from pydicom.valuerep import VR

from wsidicom.metadata.schema.dicom.fields import (
    DateDicomField,
    DefaultingNoneDicomField,
    DefaultingTagDicomField,
    PersonNameDicomField,
    StringDicomField,
    TimeDicomField,
    UidDicomField,
)
from wsidicom.metadata.schema.dicom.schema import ModuleDicomSchema
from wsidicom.metadata.study import Study


class StudyDicomSchema(ModuleDicomSchema[Study]):
    uid = DefaultingTagDicomField(
        UidDicomField(), tag="default_uid", data_key="StudyInstanceUID", allow_none=True
    )
    identifier = DefaultingNoneDicomField(
        StringDicomField(value_representation=VR.ST),
        data_key="StudyID",
        allow_none=True,
    )
    date = DefaultingNoneDicomField(
        DateDicomField(), data_key="StudyDate", allow_none=True
    )
    time = DefaultingNoneDicomField(
        TimeDicomField(), data_key="StudyTime", allow_none=True
    )
    accession_number = DefaultingNoneDicomField(
        StringDicomField(value_representation=VR.ST),
        data_key="AccessionNumber",
        allow_none=True,
    )
    referring_physician_name = DefaultingNoneDicomField(
        PersonNameDicomField(), data_key="ReferringPhysicianName", allow_none=True
    )

    @property
    def load_type(self) -> Type[Study]:
        return Study

    @property
    def module_name(self) -> str:
        return "study"



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Equipment model."""

from dataclasses import dataclass, replace
from typing import Optional, Sequence


@dataclass(frozen=True)
class Equipment:
    """
    Equipment used to produce the slide.

    Corresponds to the `Required` attributes in the Enhanced General Equipment Module:
    https://dicom.nema.org/medical/dicom/current/output/chtml/part03/sect_C.7.5.2.html

    Parameters
    ----------
    manufacturer : Optional[str] = None
        The scanner manufacturer.
    model_name : Optional[str] = None
        The scanner model name.
    device_serial_number : Optional[str] = None
        The scanner device serial number.
    software_versions : Optional[Sequence[str]] = None
        The scanner software versions.
    """

    manufacturer: Optional[str] = None
    model_name: Optional[str] = None
    device_serial_number: Optional[str] = None
    software_versions: Optional[Sequence[str]] = None

    def remove_confidential(self) -> "Equipment":
        return replace(
            self,
            device_serial_number=None,
        )



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Schemas for serializing specimen description."""

import datetime
import logging
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Optional,
    Sequence,
    Type,
    Union,
)

from marshmallow import ValidationError, fields, post_load
from pydicom import Dataset
from pydicom.sr.codedict import codes
from pydicom.sr.coding import Code
from pydicom.valuerep import VR

from wsidicom.conceptcode import (
    AnatomicPathologySpecimenTypesCode,
    ContainerTypeCode,
    SpecimenCollectionProcedureCode,
    SpecimenEmbeddingMediaCode,
    SpecimenFixativesCode,
    SpecimenPreparationProcedureCode,
    SpecimenPreparationStepsCode,
    SpecimenSamplingProcedureCode,
    SpecimenStainsCode,
    dataset_to_code,
)
from wsidicom.config import settings
from wsidicom.metadata.sample import Measurement, SampleLocalization
from wsidicom.metadata.schema.common import DefaultOnValidationExceptionField
from wsidicom.metadata.schema.dicom.fields import (
    CodeDicomField,
    CodeItemDicomField,
    DateTimeItemDicomField,
    IssuerOfIdentifierDicomField,
    MeasurementtemDicomField,
    SingleCodeSequenceField,
    StringDicomField,
    StringItemDicomField,
    UidDicomField,
)
from wsidicom.metadata.schema.dicom.sample.model import (
    CollectionDicomModel,
    ProcessingDicomModel,
    ReceivingDicomModel,
    SamplingDicomModel,
    SpecimenDescriptionDicomModel,
    SpecimenPreparationStepDicomModel,
    StainingDicomModel,
    StorageDicomModel,
)
from wsidicom.metadata.schema.dicom.schema import (
    DicomSchema,
    ItemField,
    ItemSequenceDicomSchema,
    LoadType,
)


class SampleCodes:
    identifier: Code = codes.DCM.SpecimenIdentifier  # type: ignore
    issuer_of_identifier: Code = codes.DCM.IssuerOfSpecimenIdentifier  # type: ignore
    processing_type: Code = codes.DCM.ProcessingType  # type: ignore
    sampling_method: Code = codes.DCM.SamplingMethod  # type: ignore
    datetime_of_processing: Code = codes.DCM.DatetimeOfProcessing  # type: ignore
    processing_description: Code = codes.DCM.ProcessingStepDescription  # type: ignore
    parent_specimen_identifier: Code = codes.DCM.ParentSpecimenIdentifier  # type: ignore
    issuer_of_parent_specimen_identifier: Code = codes.DCM.IssuerOfParentSpecimenIdentifier  # type: ignore
    parent_specimen_type: Code = codes.DCM.ParentSpecimenType  # type: ignore
    specimen_type: Code = codes.SCT.SpecimenType  # type: ignore
    specimen_collection: Code = codes.SCT.SpecimenCollection  # type: ignore
    sampling_of_tissue_specimen: Code = codes.SCT.SamplingOfTissueSpecimen  # type: ignore
    sample_processing: Code = codes.SCT.SpecimenProcessing  # type: ignore
    staining: Code = codes.SCT.Staining  # type: ignore
    using_substance: Code = codes.SCT.UsingSubstance  # type: ignore
    fixative: Code = codes.SCT.TissueFixative  # type: ignore
    embedding: Code = codes.SCT.TissueEmbeddingMedium  # type: ignore
    location_frame_of_reference: Code = codes.DCM.PositionFrameOfReference  # type: ignore
    location_of_sampling_site: Code = codes.DCM.LocationOfSamplingSite  # type: ignore
    location_of_sampling_site_x: Code = codes.DCM.LocationOfSamplingSiteXOffset  # type: ignore
    location_of_sampling_site_y: Code = codes.DCM.LocationOfSamplingSiteYOffset  # type: ignore
    location_of_sampling_site_z: Code = codes.DCM.LocationOfSamplingSiteZOffset  # type: ignore
    location_of_specimen: Code = codes.DCM.LocationOfSpecimen  # type: ignore
    location_of_specimen_x: Code = codes.DCM.LocationOfSpecimenXOffset  # type: ignore
    location_of_specimen_y: Code = codes.DCM.LocationOfSpecimenYOffset  # type: ignore
    location_of_specimen_z: Code = codes.DCM.LocationOfSpecimenZOffset  # type: ignore
    visual_marking_of_specimen: Code = codes.DCM.VisualMarkingOfSpecimen  # type: ignore
    container: Code = codes.SCT.SpecimenContainer  # type: ignore
    receiving: Code = codes.SCT.SpecimenReceiving  # type: ignore
    storage: Code = codes.DCM.SpecimenStorage  # type: ignore


class SampleLocalizationDicomSchema(ItemSequenceDicomSchema[SampleLocalization]):
    reference = StringItemDicomField(allow_none=True)
    description = StringItemDicomField(allow_none=True)
    x = MeasurementtemDicomField(allow_none=True)
    y = MeasurementtemDicomField(allow_none=True)
    z = MeasurementtemDicomField(allow_none=True)
    visual_marking = StringItemDicomField(allow_none=True)

    @property
    def load_type(self):
        return SampleLocalization

    @property
    def item_fields(self) -> Dict[str, ItemField]:
        return {
            "reference": ItemField(
                SampleCodes.location_frame_of_reference, (str,), False
            ),
            "description": ItemField(SampleCodes.location_of_specimen, (str,), False),
            "x": ItemField(SampleCodes.location_of_specimen_x, (Measurement,), False),
            "y": ItemField(SampleCodes.location_of_specimen_y, (Measurement,), False),
            "z": ItemField(SampleCodes.location_of_specimen_z, (Measurement,), False),
            "visual_marking": ItemField(
                SampleCodes.visual_marking_of_specimen, (str,), False
            ),
        }


class BasePreparationStepDicomSchema(ItemSequenceDicomSchema[LoadType]):
    _dump_only_fields = ["processing_type"]

    identifier = StringItemDicomField()
    issuer_of_identifier = StringItemDicomField(allow_none=True, load_default=None)
    date_time = DateTimeItemDicomField(allow_none=True, load_default=None)
    description = StringItemDicomField(allow_none=True, load_default=None)
    fixative = CodeItemDicomField(
        load_type=SpecimenFixativesCode, allow_none=True, load_default=None
    )
    embedding = CodeItemDicomField(
        load_type=SpecimenEmbeddingMediaCode, allow_none=True, load_default=None
    )
    processing = CodeItemDicomField(
        load_type=SpecimenPreparationStepsCode, allow_none=True, load_default=None
    )
    container = CodeItemDicomField(
        load_type=ContainerTypeCode, allow_none=True, load_default=None
    )
    specimen_type = CodeItemDicomField(
        load_type=AnatomicPathologySpecimenTypesCode,
        allow_none=True,
        load_default=None,
    )

    @property
    def item_fields(self) -> Dict[str, ItemField]:
        """TID 8001 Specimen Preparation, excluding Collection, Sampling, and Specimen
        fields."""
        return {
            "identifier": ItemField(SampleCodes.identifier, (str,), False),
            "issuer_of_identifier": ItemField(
                SampleCodes.issuer_of_identifier, (str,), False
            ),
            "container": ItemField(SampleCodes.container, (Code,), False),
            "specimen_type": ItemField(SampleCodes.specimen_type, (Code,), False),
            "processing_type": ItemField(SampleCodes.processing_type, (Code,), False),
            "date_time": ItemField(
                SampleCodes.datetime_of_processing,
                (datetime.datetime,),
                False,
            ),
            "description": ItemField(SampleCodes.processing_description, (str,), False),
            "processing": ItemField(SampleCodes.processing_description, (Code,), False),
            "fixative": ItemField(SampleCodes.fixative, (Code,), False),
            "embedding": ItemField(SampleCodes.embedding, (Code,), False),
        }


class SamplingDicomSchema(BasePreparationStepDicomSchema[SamplingDicomModel]):
    processing_type = CodeItemDicomField(
        load_type=SpecimenPreparationProcedureCode,
        dump_default=SampleCodes.sampling_of_tissue_specimen,
        dump_only=True,
    )
    method = CodeItemDicomField(SpecimenSamplingProcedureCode)
    parent_specimen_identifier = StringItemDicomField()
    issuer_of_parent_specimen_identifier = StringItemDicomField(
        allow_none=True,
        load_default=None,
    )
    parent_specimen_type = CodeItemDicomField(AnatomicPathologySpecimenTypesCode)
    location_reference = StringItemDicomField(allow_none=True)
    location_description = StringItemDicomField(allow_none=True)
    location_x = MeasurementtemDicomField(allow_none=True)
    location_y = MeasurementtemDicomField(allow_none=True)
    location_z = MeasurementtemDicomField(allow_none=True)

    @property
    def load_type(self):
        return SamplingDicomModel

    @property
    def item_fields(self) -> Dict[str, ItemField]:
        return {
            "identifier": ItemField(SampleCodes.identifier, (str,), False),
            "issuer_of_identifier": ItemField(
                SampleCodes.issuer_of_identifier, (str,), False
            ),
            "container": ItemField(SampleCodes.container, (Code,), False),
            "specimen_type": ItemField(SampleCodes.specimen_type, (Code,), False),
            "processing_type": ItemField(SampleCodes.processing_type, (Code,), False),
            "date_time": ItemField(
                SampleCodes.datetime_of_processing,
                (datetime.datetime,),
                False,
            ),
            "description": ItemField(SampleCodes.processing_description, (str,), False),
            "processing": ItemField(SampleCodes.processing_description, (Code,), False),
            "method": ItemField(SampleCodes.sampling_method, (Code,), False),
            "parent_specimen_identifier": ItemField(
                SampleCodes.parent_specimen_identifier,
                (str,),
                False,
            ),
            "issuer_of_parent_specimen_identifier": ItemField(
                SampleCodes.issuer_of_parent_specimen_identifier,
                (str,),
                False,
            ),
            "parent_specimen_type": ItemField(
                SampleCodes.parent_specimen_type, (Code,), False
            ),
            "location_reference": ItemField(
                SampleCodes.location_frame_of_reference, (str,), False
            ),
            "location_description": ItemField(
                SampleCodes.location_of_sampling_site, (str,), False
            ),
            "location_x": ItemField(
                SampleCodes.location_of_sampling_site_x, (Measurement,), False
            ),
            "location_y": ItemField(
                SampleCodes.location_of_sampling_site_y, (Measurement,), False
            ),
            "location_z": ItemField(
                SampleCodes.location_of_sampling_site_z, (Measurement,), False
            ),
            "fixative": ItemField(SampleCodes.fixative, (Code,), False),
            "embedding": ItemField(SampleCodes.embedding, (Code,), False),
        }


class CollectionDicomSchema(BasePreparationStepDicomSchema[CollectionDicomModel]):
    processing_type = CodeItemDicomField(
        load_type=SpecimenPreparationProcedureCode,
        dump_default=SampleCodes.specimen_collection,
        dump_only=True,
    )
    method = CodeItemDicomField(SpecimenCollectionProcedureCode)

    @property
    def load_type(self):
        return CollectionDicomModel

    @property
    def item_fields(self) -> Dict[str, ItemField]:
        return {
            "identifier": ItemField(SampleCodes.identifier, (str,), False),
            "issuer_of_identifier": ItemField(
                SampleCodes.issuer_of_identifier, (str,), False
            ),
            "container": ItemField(SampleCodes.container, (Code,), False),
            "specimen_type": ItemField(SampleCodes.specimen_type, (Code,), False),
            "processing_type": ItemField(SampleCodes.processing_type, (Code,), False),
            "date_time": ItemField(
                SampleCodes.datetime_of_processing,
                (datetime.datetime,),
                False,
            ),
            "description": ItemField(SampleCodes.processing_description, (str,), False),
            "processing": ItemField(SampleCodes.processing_description, (Code,), False),
            "method": ItemField(SampleCodes.specimen_collection, (Code,), False),
            "fixative": ItemField(SampleCodes.fixative, (Code,), False),
            "embedding": ItemField(SampleCodes.embedding, (Code,), False),
        }


class ProcessingDicomSchema(BasePreparationStepDicomSchema[ProcessingDicomModel]):
    processing_type = CodeItemDicomField(
        load_type=SpecimenPreparationProcedureCode,
        dump_default=SampleCodes.sample_processing,
        dump_only=True,
    )
    processing = CodeItemDicomField(
        load_type=SpecimenPreparationStepsCode, allow_none=True, load_default=None
    )

    @property
    def load_type(self):
        return ProcessingDicomModel


class SubstanceItemDicomField(fields.Field):
    _code_list_field = fields.List(CodeItemDicomField(SpecimenStainsCode))
    _string_field = StringItemDicomField()

    def _serialize(
        self,
        value: Optional[Union[str, Sequence[SpecimenStainsCode]]],
        attr,
        obj,
        **kwargs,
    ):
        if value is None:
            return None
        if isinstance(value, str):
            return [self._string_field._serialize(value, attr, obj, **kwargs)]
        return self._code_list_field._serialize(value, attr, obj, **kwargs)

    def _deserialize(self, datasets: Sequence[Dataset], attr, obj, **kwargs):
        first_value = datasets[0]

        if first_value.ValueType == "TEXT" or hasattr(first_value, "TextValue"):
            return self._string_field.deserialize(first_value, attr, obj, **kwargs)
        return self._code_list_field.deserialize(datasets, attr, obj, **kwargs)


class StainingDicomSchema(BasePreparationStepDicomSchema[StainingDicomModel]):
    processing_type = CodeItemDicomField(
        load_type=SpecimenPreparationProcedureCode,
        dump_default=SampleCodes.staining,
        dump_only=True,
    )
    substances = SubstanceItemDicomField()

    @property
    def load_type(self):
        return StainingDicomModel

    @property
    def item_fields(self) -> Dict[str, ItemField]:
        return {
            "identifier": ItemField(SampleCodes.identifier, (str,), False),
            "issuer_of_identifier": ItemField(
                SampleCodes.issuer_of_identifier, (str,), False
            ),
            "container": ItemField(SampleCodes.container, (Code,), False),
            "specimen_type": ItemField(SampleCodes.specimen_type, (Code,), False),
            "processing_type": ItemField(SampleCodes.processing_type, (Code,), False),
            "date_time": ItemField(
                SampleCodes.datetime_of_processing,
                (datetime.datetime,),
                False,
            ),
            "description": ItemField(SampleCodes.processing_description, (str,), False),
            "processing": ItemField(SampleCodes.processing_description, (Code,), False),
            "substances": ItemField(SampleCodes.using_substance, (str, Code), True),
            "fixative": ItemField(SampleCodes.fixative, (Code,), False),
            "embedding": ItemField(SampleCodes.embedding, (Code,), False),
        }


class ReceivingDicomSchema(BasePreparationStepDicomSchema[ReceivingDicomModel]):
    processing_type = CodeItemDicomField(
        load_type=SpecimenPreparationProcedureCode,
        dump_default=SampleCodes.receiving,
        dump_only=True,
    )

    @property
    def load_type(self):
        return ReceivingDicomModel


class StorageDicomSchema(BasePreparationStepDicomSchema[StorageDicomModel]):
    processing_type = CodeItemDicomField(
        load_type=SpecimenPreparationProcedureCode,
        dump_default=SampleCodes.storage,
        dump_only=True,
    )

    @property
    def load_type(self):
        return StorageDicomModel


class PreparationStepDicomField(fields.Field):
    """Mapping step type to schema."""

    _type_to_schema_mapping: Dict[
        Type[SpecimenPreparationStepDicomModel], Type[ItemSequenceDicomSchema]
    ] = {
        SamplingDicomModel: SamplingDicomSchema,
        CollectionDicomModel: CollectionDicomSchema,
        ProcessingDicomModel: ProcessingDicomSchema,
        StainingDicomModel: StainingDicomSchema,
        ReceivingDicomModel: ReceivingDicomSchema,
        StorageDicomModel: StorageDicomSchema,
    }

    """Mapping key in serialized step to schema. Keys are CID 8111 codes."""
    _processing_type_to_schema_mapping: Dict[Code, Type[ItemSequenceDicomSchema]] = {
        SampleCodes.sampling_of_tissue_specimen: SamplingDicomSchema,
        SampleCodes.specimen_collection: CollectionDicomSchema,
        SampleCodes.sample_processing: ProcessingDicomSchema,
        SampleCodes.staining: StainingDicomSchema,
        SampleCodes.receiving: ReceivingDicomSchema,
        SampleCodes.storage: StorageDicomSchema,
    }

    def _serialize(
        self,
        step: SpecimenPreparationStepDicomModel,
        attr: Optional[str],
        obj: Any,
        **kwargs,
    ) -> Dataset:
        """Serialize step to dataset."""
        assert self.data_key is not None
        sequence = self._subschema_dump(step)
        dataset = Dataset()
        setattr(dataset, self.data_key, sequence)
        return dataset

    def _deserialize(
        self, dataset: Dataset, attr: Optional[str], data: Any, **kwargs
    ) -> Optional[SpecimenPreparationStepDicomModel]:
        """Deserialize step from dataset."""
        assert self.data_key is not None
        sequence = getattr(dataset, self.data_key)
        return self._subschema_load(sequence)

    def _subschema_load(
        self, sequence: Iterable[Dataset]
    ) -> Optional[SpecimenPreparationStepDicomModel]:
        """Select a schema and load and return step using the schema."""

        def compare_codes(code: Code, other: Code) -> bool:
            # Code comparision do mapping from SRT to SCT codes, but if the code is not
            # in the mapping it will raise KeyError.
            # See https://github.com/pydicom/pydicom/issues/1994
            try:
                return code == other
            except KeyError:
                return False

        try:
            try:
                processing_type: Code = next(
                    dataset_to_code(item.ConceptCodeSequence[0])
                    for item in sequence
                    if compare_codes(
                        dataset_to_code(item.ConceptNameCodeSequence[0]),
                        SampleCodes.processing_type,
                    )
                )
            except StopIteration:
                raise ValidationError(
                    "Failed to load processing step due to missing processing type."
                )
            try:
                # Iterate over schemas as code comparison can map SRT to SCT codes
                schema = next(
                    schema
                    for name, schema in self._processing_type_to_schema_mapping.items()
                    if compare_codes(name, processing_type)
                )
            except StopIteration:
                raise ValidationError(
                    "Failed to load processing step due to unknown "
                    f"processing type {processing_type}."
                )
            loaded = schema().load(sequence, many=False)
        except ValidationError as exception:
            error = "Failed to load processing step due to validation error."
            if settings.ignore_specimen_preparation_step_on_validation_error:
                logging.warning(error, exc_info=True)
                return None
            raise ValidationError(
                error + " Check the processing step for errors or set "
                "`settings.ignore_specimen_preparation_step_on_validation_error` "
                "to `True` to ignore this step."
            ) from exception
        assert isinstance(loaded, SpecimenPreparationStepDicomModel)
        return loaded

    def _subschema_dump(self, step: SpecimenPreparationStepDicomModel) -> List[Dataset]:
        """Select a schema and dump the step using the schema."""
        schema = self._type_to_schema_mapping[type(step)]
        dumped = schema().dump(step, many=False)
        assert isinstance(dumped, list)
        return dumped


class SpecimenDescriptionDicomSchema(DicomSchema[SpecimenDescriptionDicomModel]):
    identifier = StringDicomField(
        value_representation=VR.LO, data_key="SpecimenIdentifier"
    )
    uid = UidDicomField(data_key="SpecimenUID")
    localization = fields.Nested(
        SampleLocalizationDicomSchema(),
        data_key="SpecimenLocalizationContentItemSequence",
        allow_none=True,
    )
    issuer_of_identifier = IssuerOfIdentifierDicomField(
        data_key="IssuerOfTheSpecimenIdentifierSequence", allow_none=True
    )
    steps = DefaultOnValidationExceptionField(
        fields.List(
            PreparationStepDicomField(
                data_key="SpecimenPreparationStepContentItemSequence"
            )
        ),
        data_key="SpecimenPreparationSequence",
        load_default=[],
    )
    anatomical_sites = fields.List(
        CodeDicomField(Code),
        data_key="PrimaryAnatomicStructureSequence",
        load_default=[],
    )
    specimen_type = SingleCodeSequenceField(
        load_type=AnatomicPathologySpecimenTypesCode,
        data_key="SpecimenTypeCodeSequence",
        dump_default=AnatomicPathologySpecimenTypesCode("Slide"),
        dump_only=True,
    )
    short_description = StringDicomField(
        value_representation=VR.LO, data_key="SpecimenShortDescription", allow_none=True
    )
    detailed_description = StringDicomField(
        value_representation=VR.UT,
        data_key="SpecimenDetailedDescription",
        allow_none=True,
    )

    @property
    def load_type(self):
        return SpecimenDescriptionDicomModel

    @post_load
    def post_load(
        self, data: Dict[str, Any], **kwargs
    ) -> SpecimenDescriptionDicomModel:
        """Remove None values from steps before loading to object."""
        data["steps"] = [step for step in data["steps"] if step is not None]
        return super().post_load(data, **kwargs)



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Label model."""

from dataclasses import dataclass, replace
from typing import List, Optional


@dataclass(frozen=True)
class Label:
    """
    Label metadata.

    Corresponds to the `Required, Empty if Unknown` attributes in the Slide Label
    module:
    https://dicom.nema.org/medical/dicom/current/output/chtml/part03/sect_C.8.12.8.html

    Parameters
    ----------
    text : Optional[str] = None
        The label text.
    barcode : Optional[str] = None
        The label barcode.
    label_in_volume_image : bool = False
        Whether the label is present in the volume image.
    label_in_overview_image : bool = False
        Whether the label is present in the overview image.
    label_is_phi : bool = True
        Whether the label is personal health information.
    """

    text: Optional[str] = None
    barcode: Optional[str] = None
    label_in_volume_image: bool = False
    label_in_overview_image: bool = False
    label_is_phi: bool = True

    @classmethod
    def merge_image_types(
        cls, volume: "Label", label: Optional["Label"], overview: Optional["Label"]
    ):
        labels: List[Optional[Label]] = [label, volume, overview]
        text = next((item.text for item in labels if item is not None), None)
        barcode = next((item.barcode for item in labels if item is not None), None)
        label_is_phi = any(item.label_is_phi for item in labels if item is not None)
        if overview is None:
            label_in_overview_image = False
        else:
            label_in_overview_image = overview.label_in_overview_image
        return cls(
            text=text,
            barcode=barcode,
            label_in_volume_image=volume.label_in_volume_image,
            label_in_overview_image=label_in_overview_image,
            label_is_phi=label_is_phi,
        )

    def remove_confidential(self) -> "Label":
        return replace(self, text=None, barcode=None)



#    Copyright 2021, 2022, 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Abstract class for FrameIndex that has an offset table (basic or extended)."""

from abc import abstractmethod
from struct import unpack
from typing import List, Tuple

from pydicom.tag import ItemTag

from wsidicom.errors import WsiDicomFileError
from wsidicom.file.io.frame_index.encapsulated_pixel_data import EncapsulatedPixelData


class OffsetTable(EncapsulatedPixelData):
    @property
    @abstractmethod
    def bytes_per_item(self) -> int:
        """Return the number of bytes per item in the table."""
        raise NotImplementedError()

    @property
    @abstractmethod
    def mode(self) -> str:
        """Return the mode used for unpacking the table."""
        raise NotImplementedError()

    def _parse_table(self, table: bytes, pixels_start: int) -> List[Tuple[int, int]]:
        """Parse table with offsets (BOT or EOT).

        Parameters
        ----------
        table: bytes
            BOT or EOT as bytes
        table_type: OffsetTableType
            Type of table, 'bot' or 'eot'.
        pixels_start: int
            Position of first frame item in pixel data.

        Returns
        -------
        List[Tuple[int, int]]
            A list with frame positions and frame lengths.
        """
        if not self._file.is_little_endian:
            raise WsiDicomFileError(
                str(self._file), "Big endian not supported for BOT or EOT"
            )
        bytes_per_item = self.bytes_per_item
        mode = self.mode
        table_length = len(table)
        TAG_BYTES = 4
        LENGTH_BYTES = 4
        positions: List[Tuple[int, int]] = []
        # Read through table to get offset and length for all but last item
        # All read offsets are for item tag of frame and relative to first
        # frame in pixel data.
        this_offset: int = unpack(mode, table[0:bytes_per_item])[0]
        if this_offset != 0:
            raise ValueError("First item in table should be at offset 0")
        for index in range(bytes_per_item, table_length, bytes_per_item):
            next_offset = unpack(mode, table[index : index + bytes_per_item])[0]
            offset = this_offset + TAG_BYTES + LENGTH_BYTES
            length = next_offset - offset
            if length <= 0 or length % 2:
                raise WsiDicomFileError(
                    str(self._file),
                    (
                        f"Invalid frame length {length} "
                        f"for frame {index // bytes_per_item}"
                    ),
                )
            positions.append((pixels_start + offset, length))
            this_offset = next_offset

        # Go to last frame in pixel data and read the length of the frame
        self._file.seek(pixels_start + this_offset)
        if self._file.read_le_tag() != ItemTag:
            raise WsiDicomFileError(str(self._file), "Expected ItemTag in PixelData")
        length: int = self._file.read_leUL()
        if length <= 0 or length % 2:
            raise WsiDicomFileError(str(self._file), "Invalid frame length")
        offset = this_offset + TAG_BYTES + LENGTH_BYTES
        positions.append((pixels_start + offset, length))

        return positions



#    Copyright 2021, 2022, 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from functools import cached_property, lru_cache
from typing import List, Optional, Sequence, Set

from wsidicom.errors import WsiDicomNotFoundError
from wsidicom.geometry import Point
from wsidicom.instance.dataset import WsiDataset
from wsidicom.instance.tile_index.tile_index import TileIndex


class FullTileIndex(TileIndex):
    """Index for mapping tile position to frame number for datasets containing
    full tiles. Pixel data tiles are ordered by column, row, z and path, thus
    the frame index for a tile can directly be calculated."""

    def __init__(self, datasets: Sequence[WsiDataset]):
        """Create full tile index for frames in datasets. Requires equal tile
        size for all tile planes.

        Parameters
        ----------
        datasets: Sequence[Dataset]
            List of datasets containing full tiled image data.
        """
        super().__init__(datasets)

    @cached_property
    def focal_planes(self) -> List[float]:
        return self._read_focal_planes_from_datasets()

    def __str__(self) -> str:
        return self.pretty_str()

    def pretty_str(self, indent: int = 0, depth: Optional[int] = None) -> str:
        string = (
            f"Full tile index tile size: {self.tile_size}"
            f", plane size: {self.tiled_size}"
        )
        if depth is not None:
            depth -= 1
            if depth < 0:
                return string
        string += f" of z: {self.focal_planes} and path: {self.optical_paths}"

        return string

    def get_frame_index(self, tile: Point, z: float, path: str) -> int:
        """Return frame index for a Point tile, z coordinate, and optical path
        from full tile index. Assumes that tile, z, and path are valid.

        Parameters
        ----------
        tile: Point
            Tile xy to get.
        z: float
            Z coordinate to get.
        path: str
            ID of optical path to get.

        Returns
        -------
        int
            Frame index.
        """
        plane_offset = tile.x + self.tiled_size.width * tile.y
        z_offset = self._get_focal_plane_index(z) * self.tiled_size.area
        path_offset = (
            self._get_optical_path_index(path)
            * len(self.focal_planes)
            * self.tiled_size.area
        )
        return plane_offset + z_offset + path_offset

    def _read_focal_planes_from_datasets(
        self,
    ) -> List[float]:
        """Return list of focal planes in datasets. Values in Pixel Measures
        Sequence are in mm.

        Returns
        -------
        List[float]
            Focal planes, specified in um.

        """
        MM_TO_MICRON = 1000.0
        DECIMALS = 3
        focal_planes: Set[float] = set()
        for dataset in self._datasets:
            slice_spacing = dataset.spacing_between_slices
            number_of_focal_planes = getattr(dataset, "TotalPixelMatrixFocalPlanes", 1)
            if slice_spacing is None:
                if number_of_focal_planes == 1:
                    slice_spacing = 0.0
                else:
                    raise ValueError(
                        "Slice spacing must be known if multiple focal planes."
                    )
            elif slice_spacing == 0 and number_of_focal_planes != 1:
                raise ValueError(
                    "Slice spacing must be non-zero if multiple focal planes."
                )

            try:
                z_offset = (
                    dataset.SharedFunctionalGroupsSequence[0]
                    .PlanePositionSlideSequence[0]
                    .ZOffsetInSlideCoordinateSystem
                )
            except AttributeError:
                z_offset = 0

            for plane in range(number_of_focal_planes):
                z = z_offset + round(plane * slice_spacing * MM_TO_MICRON, DECIMALS)
                focal_planes.add(z)
        return sorted(list(focal_planes))

    @lru_cache
    def _get_optical_path_index(self, path: str) -> int:
        """Return index of the optical path in instance.
        This assumes that all files in a concatenated set contains all the
        optical path identifiers of the set.

        Parameters
        ----------
        path: str
            Optical path identifier to search for.

        Returns
        -------
        int
            The index of the optical path identifier in the optical path
            sequence.
        """
        try:
            return next(
                (
                    index
                    for index, plane_path in enumerate(self._optical_paths)
                    if plane_path == path
                )
            )
        except StopIteration:
            raise WsiDicomNotFoundError(f"Optical path {path}", str(self))

    @lru_cache
    def _get_focal_plane_index(self, z: float) -> int:
        """Return index of the focal plane of z.

        Parameters
        ----------
        z: float
            The z coordinate (in um) to search for.

        Returns
        -------
        int
            Focal plane index for z coordinate.
        """
        try:
            return next(
                index for index, plane in enumerate(self.focal_planes) if plane == z
            )
        except StopIteration:
            raise WsiDicomNotFoundError(f"Z {z} in instance", str(self))



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Patient model."""

import datetime
from dataclasses import dataclass, replace
from enum import Enum
from typing import Optional, Sequence, Union

from pydicom.sr.coding import Code


class PatientSex(Enum):
    """Patient sex."""

    F = "female"
    M = "male"
    O = "other"  # noqa: E741


@dataclass(frozen=True)
class PatientDeIdentification:
    """Patient de-identification.

    Parameters
    ----------
    identity_removed : bool
        Whether the patient identity has been removed.
    methods : Optional[Sequence[Union[str, Code]]] = None
        The methods used to de-identify the patient.
    """

    identity_removed: bool
    methods: Optional[Sequence[Union[str, Code]]] = None


@dataclass(frozen=True)
class Patient:
    """
    Patient metadata.

    Corresponds to the `Required` and `Required, Empty if Unknown` attributes in the
    Patient Module:
    https://dicom.nema.org/medical/dicom/current/output/chtml/part03/sect_C.7.html

    Parameters
    ----------
    name : Optional[str] = None
        The patient name.
    identifier : Optional[str] = None
        The patient identifier.
    birth_date : Optional[datetime.date] = None
        The patient birth date.
    sex : Optional[PatientSex] = None
        The sex of the patient.
    species_description : Optional[Union[str, Code]] = None
        The species description of the patient.
    de_identification : Optional[PatientDeIdentification] = None
        The de-identification of the patient.
    """

    name: Optional[str] = None
    identifier: Optional[str] = None
    birth_date: Optional[datetime.date] = None
    sex: Optional[PatientSex] = None
    species_description: Optional[Union[str, Code]] = None
    de_identification: Optional[PatientDeIdentification] = None

    def remove_confidential(self) -> "Patient":
        return replace(
            self,
            name=None,
            identifier=None,
            birth_date=None,
            sex=None,
            de_identification=(
                replace(self.de_identification, identity_removed=True)
                if self.de_identification
                else None
            ),
        )



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Module for serializing specimen description."""

from wsidicom.metadata.schema.dicom.sample.model import SpecimenDescriptionDicomModel
from wsidicom.metadata.schema.dicom.sample.schema import SpecimenDescriptionDicomSchema

__all__ = ["SpecimenDescriptionDicomModel", "SpecimenDescriptionDicomSchema"]



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Schemas for serializing metadata models to and from DICOM."""

from wsidicom.metadata.schema.dicom.equipment import EquipmentDicomSchema
from wsidicom.metadata.schema.dicom.image import ImageDicomSchema
from wsidicom.metadata.schema.dicom.label import LabelDicomSchema
from wsidicom.metadata.schema.dicom.optical_path import OpticalPathDicomSchema
from wsidicom.metadata.schema.dicom.patient import PatientDicomSchema
from wsidicom.metadata.schema.dicom.series import SeriesDicomSchema
from wsidicom.metadata.schema.dicom.slide import SlideDicomSchema
from wsidicom.metadata.schema.dicom.study import StudyDicomSchema
from wsidicom.metadata.schema.dicom.wsi import WsiMetadataDicomSchema

__all__ = [
    "EquipmentDicomSchema",
    "ImageDicomSchema",
    "LabelDicomSchema",
    "OpticalPathDicomSchema",
    "PatientDicomSchema",
    "SeriesDicomSchema",
    "SlideDicomSchema",
    "StudyDicomSchema",
    "WsiMetadataDicomSchema",
]



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""DICOM schema for Label model."""

from typing import Any, Dict, Type

from marshmallow import fields, post_dump, post_load, pre_dump
from pydicom.valuerep import VR

from wsidicom.instance import ImageType
from wsidicom.metadata.label import Label
from wsidicom.metadata.schema.dicom.fields import (
    BooleanDicomField,
    DefaultingDicomField,
    StringDicomField,
)
from wsidicom.metadata.schema.dicom.schema import ModuleDicomSchema


class LabelDicomSchema(ModuleDicomSchema[Label]):
    text = DefaultingDicomField(
        StringDicomField(VR.UT),
        data_key="LabelText",
        allow_none=True,
        dump_default=None,
    )
    barcode = DefaultingDicomField(
        StringDicomField(VR.LO),
        data_key="BarcodeValue",
        allow_none=True,
        dump_default=None,
    )

    label_in_volume_image = BooleanDicomField(load_only=True, allow_none=True)
    label_in_overview_image = BooleanDicomField(load_only=True, allow_none=True)
    label_is_phi = BooleanDicomField(load_only=True, allow_none=True)
    burned_in_annotation = BooleanDicomField(
        data_key="BurnedInAnnotation", load_default=False
    )
    specimen_label_in_image = BooleanDicomField(
        data_key="SpecimenLabelInImage", load_default=False
    )
    image_type = fields.List(
        StringDicomField(VR.CS), load_only=True, data_key="ImageType"
    )

    @property
    def load_type(self) -> Type[Label]:
        return Label

    @pre_dump
    def pre_dump(self, label: Label, **kwargs):
        image_type = self.context.get("image_type", None)
        label_in_image = False
        contains_phi = False
        if (
            (image_type == ImageType.VOLUME and label.label_in_volume_image)
            or (image_type == ImageType.OVERVIEW and label.label_in_overview_image)
            or image_type == ImageType.LABEL
        ):
            label_in_image = "True"
            contains_phi = label.label_is_phi
        attributes = {
            "text": label.text,
            "barcode": label.barcode,
            "burned_in_annotation": contains_phi,
            "specimen_label_in_image": label_in_image,
        }

        return attributes

    @post_dump
    def post_dump(self, data: Dict[str, Any], **kwargs):
        # Remove text and barcode if empty and not label image type
        image_type = self.context.get("image_type", None)
        if image_type != ImageType.LABEL:
            if data["LabelText"] is None:
                data.pop("LabelText")
            if data["BarcodeValue"] is None:
                data.pop("BarcodeValue")

        return super().post_dump(data, **kwargs)

    @post_load
    def post_load(self, data: Dict[str, Any], **kwargs):
        image_type = ImageType(data.pop("image_type")[2])
        burned_in_annotation = data.pop("burned_in_annotation")
        specimen_label_in_image = data.pop("specimen_label_in_image")
        label_is_phi = False
        label_in_volume_image = False
        label_in_overview_image = False
        if image_type == ImageType.LABEL:
            label_is_phi = burned_in_annotation
        elif image_type == ImageType.VOLUME and specimen_label_in_image:
            label_is_phi = burned_in_annotation
            label_in_volume_image = True
        elif image_type == ImageType.OVERVIEW and specimen_label_in_image:
            label_is_phi = burned_in_annotation
            label_in_overview_image = True
        data["label_is_phi"] = label_is_phi
        data["label_in_volume_image"] = label_in_volume_image
        data["label_in_overview_image"] = label_in_overview_image
        return super().post_load(data, **kwargs)

    @property
    def module_name(self) -> str:
        return "label"



#    Copyright 2021, 2022, 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""A source for reading WSI DICOM files."""

import logging
from collections import defaultdict
from pathlib import Path
from typing import (
    Any,
    BinaryIO,
    Dict,
    Iterable,
    List,
    Optional,
    Union,
)

from pydicom.fileset import FileSet
from pydicom.uid import UID, MediaStorageDirectoryStorage
from upath import UPath

from wsidicom.errors import (
    WsiDicomNotFoundError,
    WsiDicomNotSupportedError,
)
from wsidicom.file.io import WsiDicomReader
from wsidicom.file.io.wsidicom_io import WsiDicomIO
from wsidicom.file.wsidicom_file_image_data import WsiDicomFileImageData
from wsidicom.file.wsidicom_stream_opener import WsiDicomStreamOpener
from wsidicom.geometry import Size
from wsidicom.graphical_annotations import AnnotationInstance
from wsidicom.instance import ImageType, TileType, WsiDataset, WsiInstance
from wsidicom.source import Source
from wsidicom.uid import ANN_SOP_CLASS_UID, WSI_SOP_CLASS_UID, SlideUids


class WsiDicomFileSource(Source):
    """Source reading WSI DICOM file instances."""

    def __init__(self, streams: Iterable[WsiDicomIO]) -> None:
        """Create a WsiDicomFileSource.

        Parameters
        ----------
        streams: Iterable[WsiDicomIO]
            Opened streams to read from.
        """
        self._levels: List[WsiDicomReader] = []
        self._labels: List[WsiDicomReader] = []
        self._overviews: List[WsiDicomReader] = []
        self._annotations: List[WsiDicomIO] = []
        for stream in streams:
            try:
                if stream.media_storage_sop_class_uid == WSI_SOP_CLASS_UID:
                    try:
                        reader = WsiDicomReader(stream)
                        if reader.image_type == ImageType.VOLUME:
                            self._levels.append(reader)
                        elif reader.image_type == ImageType.LABEL:
                            self._labels.append(reader)
                        elif reader.image_type == ImageType.OVERVIEW:
                            self._overviews.append(reader)
                    except WsiDicomNotSupportedError:
                        logging.info(f"Non-supported file {stream}.")
                        if stream.owned:
                            stream.close()
                elif stream.media_storage_sop_class_uid == ANN_SOP_CLASS_UID:
                    self._annotations.append(stream)
            except Exception:
                logging.error(
                    f"Failed to open file {stream} due to exception.", exc_info=True
                )
                stream.close()
        if len(self._levels) == 0:
            raise WsiDicomNotFoundError("Level files", "provided files")
        self._base_dataset = self._get_base_dataset(self._levels)
        self._slide_uids = self._base_dataset.uids.slide
        self._base_tile_size = self._base_dataset.tile_size

    @property
    def base_dataset(self) -> WsiDataset:
        """The dataset of the base level instance."""
        return self._base_dataset

    @property
    def level_instances(self) -> Iterable[WsiInstance]:
        """The level instances parsed from the source."""
        return self._create_instances(
            self._levels, self._slide_uids, self._base_tile_size
        )

    @property
    def label_instances(self) -> Iterable[WsiInstance]:
        """The label instances parsed from the source."""
        return self._create_instances(self._labels, self._slide_uids)

    @property
    def overview_instances(self) -> Iterable[WsiInstance]:
        """The overview instances parsed from the source."""
        return self._create_instances(self._overviews, self._slide_uids)

    @property
    def annotation_instances(self) -> Iterable[AnnotationInstance]:
        """The annotation instances parsed from the source."""
        return (
            AnnotationInstance.open_dataset(file.read_dataset())
            for file in self._annotations
        )

    @property
    def readers(self) -> List[WsiDicomReader]:
        """Return the readers in the source."""
        reader_lists: List[List[WsiDicomReader]] = [
            self._levels,
            self._labels,
            self._overviews,
        ]
        return [reader for reader_list in reader_lists for reader in reader_list]

    @property
    def files(self) -> Optional[List[UPath]]:
        """Return the files in the source, if any."""
        if all(reader.filepath is None for reader in self.readers):
            return None
        return [
            reader.filepath for reader in self.readers if reader.filepath is not None
        ]

    @property
    def is_ready_for_viewing(self) -> Optional[bool]:
        """
        Returns True if files in source are formatted for fast viewing.

        Returns None if no files are in source.
        """
        if not self.contains_levels:
            return None
        files = sorted(self.readers, key=lambda file: file.frame_count)
        for file in files:
            if file.image_type is None:
                continue
            if (
                file.dataset.tile_type != TileType.SPARSE
                or file.offset_table_type is None
            ):
                return False

        return True

    @property
    def contains_levels(self) -> bool:
        """Returns true source has one level that can be read with WsiDicom."""
        return len(self._levels) > 0

    @classmethod
    def open(
        cls,
        files: Union[str, Path, UPath, Iterable[Union[str, Path, UPath]]],
        file_options: Optional[Dict[str, Any]] = None,
    ) -> "WsiDicomFileSource":
        streams = WsiDicomStreamOpener(file_options).open(
            files, [WSI_SOP_CLASS_UID, ANN_SOP_CLASS_UID]
        )
        return cls(streams)

    @classmethod
    def open_streams(
        cls,
        streams: Iterable[BinaryIO],
    ) -> "WsiDicomFileSource":
        return cls((WsiDicomIO(stream) for stream in streams))

    @classmethod
    def open_dicomdir(
        cls,
        path: Union[str, Path, UPath],
        file_options: Optional[Dict[str, Any]] = None,
    ) -> "WsiDicomFileSource":
        """Open a DICOMDIR file and return a WsiDicomFileSource for contained files.

        Parameters
        ----------
        path: Union[str, Path, UPath]
            Path to DICOMDIR file.
        file_options: Optional[Dict[str, Any]] = None
            Keyword arguments for opening files.

        Returns
        -------
        WsiDicomFileSource
            Source for files in DICOMDIR.
        """
        files: List[str] = []
        for stream in WsiDicomStreamOpener().open(path, MediaStorageDirectoryStorage):
            dicomdir = stream.read_dataset()
            fileset = FileSet(dicomdir)
            files.extend(file.path for file in fileset)
            stream.close()
        return cls.open(files, file_options=file_options)

    def close(self) -> None:
        """Close all opened readers in the source. Does not close provided streams."""
        for reader in self.readers:
            reader.close()
        for annotation in self._annotations:
            annotation.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    @staticmethod
    def _get_base_dataset(files: Iterable[WsiDicomReader]) -> WsiDataset:
        """Return file with largest image (width) from list of files.

        Parameters
        ----------
        files: Iterable[WsiDicomReader]
           List of files.

        Returns
        -------
        WsiDataset
            Base layer dataset.
        """
        return next(
            file.dataset
            for file in sorted(
                files, reverse=True, key=lambda file: file.dataset.image_size.width
            )
        )

    @classmethod
    def _create_instances(
        cls,
        files: Iterable[WsiDicomReader],
        series_uids: SlideUids,
        series_tile_size: Optional[Size] = None,
    ) -> Iterable[WsiInstance]:
        """
        Create instances from Dicom files.

        Only files with matching series uid and tile size, if defined, are used. Other
        files are closed.

        Parameters
        ----------
        files: Iterable[WsiDicomReader]
            Files to create instances from.
        series_uids: SlideUids
            Uid to match against.
        series_tile_size: Optional[Size]
            Tile size to match against (for level instances).

        Returns
        -------
        Iterable[WsiInstancece]
            Iterable of created instances.
        """
        filtered_files = cls._filter_files(files, series_uids, series_tile_size)
        files_grouped_by_instance = cls._group_files(filtered_files)
        return (
            WsiInstance(
                [file.dataset for file in instance_files],
                WsiDicomFileImageData(instance_files),
            )
            for instance_files in files_grouped_by_instance.values()
        )

    @staticmethod
    def _filter_files(
        files: Iterable[WsiDicomReader],
        series_uids: SlideUids,
        series_tile_size: Optional[Size] = None,
    ) -> Iterable[WsiDicomReader]:
        """
        Filter list of wsi dicom files on uids and tile size if defined.

        Parameters
        ----------
        files: Iterable[WsiDicomReader]
            Wsi files to filter.
        series_uids: Uids
            Uids to check against.
        series_tile_size: Optional[Size] = None
            Tile size to check against.

        Returns
        -------
        Iterable[WsiDicomReader]
            Iterable of matching wsi dicom files.
        """
        for file in files:
            if file.dataset.matches_series(series_uids, series_tile_size):
                yield file
            else:
                logging.warning(
                    f"{file.filepath} with uids {file.uids.slide} "
                    f"did not match series with {series_uids} "
                    f"and tile size {series_tile_size}"
                )
                file.close()

    @staticmethod
    def _group_files(
        files: Iterable[WsiDicomReader],
    ) -> Dict[UID, List[WsiDicomReader]]:
        """
        Return files grouped by instance identifier (instances).

        Parameters
        ----------
        files: Iterable[WsiDicomReader]
            Files to group into instances

        Returns
        -------
        Dict[UID, List[WsiDicomReader]]
            Files grouped by instance, with instance identifier as key.
        """
        grouped_files: Dict[UID, List[WsiDicomReader]] = defaultdict(list)
        for file in files:
            grouped_files[file.uids.identifier].append(file)
        return grouped_files



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Base DICOM schemas."""

import datetime
import logging
from abc import abstractmethod
from dataclasses import dataclass
from typing import (
    Any,
    Dict,
    Generic,
    Iterable,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from marshmallow import ValidationError, post_dump, pre_load
from pydicom import Dataset
from pydicom.sr.coding import Code

from wsidicom.conceptcode import dataset_to_code
from wsidicom.metadata.sample import Measurement
from wsidicom.metadata.schema.common import LoadingSchema, LoadType
from wsidicom.metadata.schema.dicom.fields import (
    DefaultingDicomField,
    FlattenOnDumpNestedDicomField,
)

DumpType = TypeVar("DumpType", Dataset, Iterable[Dataset])


class BaseDicomSchema(LoadingSchema[LoadType], Generic[LoadType, DumpType]):
    """Base class for DICOM schemas that dumps and loads to specified type."""

    @property
    @abstractmethod
    def dump_type(self) -> Type[DumpType]:
        raise NotImplementedError()

    def load(self, dataset: DumpType, **kwargs) -> LoadType:
        """Load object from DumpType."""
        item = super().load(dataset, **kwargs)  # type: ignore
        assert isinstance(item, self.load_type)
        return item

    def dump(self, obj: LoadType, **kwargs) -> DumpType:
        """Dump object to DumpType."""
        dumped = super().dump(obj, **kwargs)
        assert isinstance(dumped, self.dump_type)
        return dumped


class DicomSchema(BaseDicomSchema[LoadType, Dataset]):
    """Base DICOM schema for attributes in a dataset."""

    @property
    def dump_type(self) -> Type[Dataset]:
        return Dataset

    @post_dump
    def post_dump(self, data: Dict[str, Any], many: bool, **kwargs) -> Dataset:
        """Create pydicom Dataset from attributes in dictionary."""
        for field in self.fields.values():
            if isinstance(field, FlattenOnDumpNestedDicomField):
                # Flatten nested fields into data
                field.flatten(data)
            if (
                field.data_key in data
                and data[field.data_key] is None
                and not isinstance(field, DefaultingDicomField)
            ):
                # Remove empty non-defaulting fields
                data.pop(field.data_key)
        dataset = Dataset()
        dataset.update(data)  # type: ignore
        return dataset

    @pre_load
    def pre_load(self, dataset: Dataset, many: bool, **kwargs) -> Dict[str, Any]:
        """Return dictionary of attributes from dataset."""
        attributes = {}
        for key, field in self.fields.items():
            if field.dump_only:
                continue
            if field.data_key is not None and field.data_key in dataset:
                attributes[field.data_key] = dataset.get(field.data_key)
            elif isinstance(field, FlattenOnDumpNestedDicomField):
                # De-flatten nested fields from dataset
                de_flattened = field.de_flatten(dataset)
                if de_flattened is not None:
                    attributes[key] = de_flattened
        return attributes


class ModuleDicomSchema(DicomSchema[LoadType]):
    """Base DICOM schema for a module, returning a default when failing to load."""

    @property
    @abstractmethod
    def module_name(self) -> str:
        raise NotImplementedError()

    def load(self, dataset: Dataset, **kwargs) -> LoadType:
        """Load dataset to LoadType. Return default LoadType if validation error."""
        try:
            return super().load(dataset, **kwargs)  # type: ignore
        except ValidationError:
            logging.warning(
                f"Failed to load module {self.module_name} with schema {self}.",
                exc_info=True,
            )
            return self.load_type()


@dataclass(frozen=True)
class ItemField:
    name: Code
    value_types: Tuple[Type, ...]
    many: bool


class ItemSequenceDicomSchema(BaseDicomSchema[LoadType, Iterable[Dataset]]):
    """Base DICOM schema for sequence of content items (each a dataset)."""

    _dump_only_fields: List[str] = []

    @property
    @abstractmethod
    def load_type(self) -> Type[LoadType]:
        raise NotImplementedError()

    @property
    def dump_type(self) -> Type[Iterable[Dataset]]:
        return list

    @property
    @abstractmethod
    def item_fields(self) -> Dict[str, ItemField]:
        """Describe the fields in the schema.

        Fields should be ordered as in TID if applicable. The key is the python name of
        the field, and the value is a ItemField with the DICOM code name of the field,
        the allowed value types (tuple of one or more types), and if the field can
        hold multiple values (e.g. is a list)."""
        raise NotImplementedError()

    @post_dump
    def post_dump(
        self, data: Dict[str, Union[Dataset, Sequence[Dataset]]], many: bool, **kwargs
    ) -> List[Dataset]:
        """Format content items into sequence in a dataset."""
        return [
            self._name_item(flatten_item, name)
            for item, name in [
                (data[key], description.name)
                for key, description in self.item_fields.items()
            ]
            if item is not None
            for flatten_item in ([item] if isinstance(item, Dataset) else item)
        ]

    @pre_load
    def pre_load(
        self, sequence: Sequence[Dataset], many: bool, **kwargs
    ) -> Dict[str, Any]:
        """Parse the sequence items from a dataset into a dictionary."""
        data = {
            key: self._get_item(sequence, description)
            for key, description in self.item_fields.items()
        }
        for field in self._dump_only_fields:
            data.pop(field)
        return data

    @staticmethod
    def _name_item(item: Dataset, name: Code):
        """Add concept name code sequence to dataset."""
        name_dataset = Dataset()
        name_dataset.CodeValue = name.value
        name_dataset.CodingSchemeDesignator = name.scheme_designator
        name_dataset.CodeMeaning = name.meaning
        if name.scheme_version is not None and name.scheme_version != "":
            name_dataset.CodingSchemeVersion = name.scheme_version
        item.ConceptNameCodeSequence = [name_dataset]
        return item

    def _get_item(
        self, sequence: Iterable[Dataset], field: ItemField
    ) -> Optional[Union[Dataset, List[Dataset]]]:
        """Get item dataset from dataset content item sequence.

        Parameters
        ----------
        dataset: Dataset
            Dataset to get item from.
        field:
            Description of the field to get.

        Returns
        -------
        Optional[Union[Dataset, List[Dataset]]]
            Item dataset or datasets or None if not found.
        """
        items = (
            item
            for item in sequence
            if dataset_to_code(item.ConceptNameCodeSequence[0]) == field.name
            and self.dataset_to_type(item) in field.value_types
        )
        if field.many:
            return list(items)
        return next(items, None)

    @staticmethod
    def dataset_to_type(dataset: Dataset) -> Type:
        value_type = dataset.ValueType
        if value_type == "CODE":
            return Code
        if value_type == "TEXT":
            return str
        if value_type == "DATETIME":
            return datetime.datetime
        if value_type == "NUMERIC":
            return Measurement
        raise NotImplementedError(
            f"Not implemented type-handling for dataset {dataset}."
        )



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from pydicom.uid import (
    RLELossless,
    ExplicitVRLittleEndian,
    UID,
    JPEGTransferSyntaxes,
    JPEGLSTransferSyntaxes,
    JPEG2000TransferSyntaxes,
)


def determine_media_type(transfer_syntax: UID) -> str:
    """Determine media type from transfer syntax.

    Parameters
    ----------
    transfer_syntax: UID
        Transfer syntax to determine media type for.

    Returns
    ----------
    str
        Media type.
    """
    if transfer_syntax in JPEGTransferSyntaxes:
        return "image/jpeg"
    if transfer_syntax in JPEGLSTransferSyntaxes:
        return "image/jls"
    if transfer_syntax in JPEG2000TransferSyntaxes:
        return "image/jp2"
    if transfer_syntax == RLELossless:
        return "image/x-dicom-rle"
    if transfer_syntax == ExplicitVRLittleEndian:
        return "application/octet-stream"
    raise NotImplementedError(f"Unsupported transfer syntax: {transfer_syntax}")



from enum import Enum
from typing import Any, Dict, List, Optional

from PIL import Image as Pillow
from PIL import UnidentifiedImageError

from wsidicom.file.io.frame_index.empty_bot import EmptyBot
from wsidicom.file.io.frame_index.offset_table_type import OffsetTableType
from wsidicom.file.io.wsidicom_io import WsiDicomIO


class EmptyTiffFrameTagsException(Exception):
    """Exception raised when file does not contain required tiff tags."""

    pass


class TiffTags(Enum):
    TILEOFFSETS = 324
    TILEBYTECOUNTS = 325


class TiffTable(EmptyBot):
    """Frame index for TIFF, parsing the index from `TileOffsets`and TileByteCounts`
    if present. Only works with `DICOM-TIFF dual files."""

    def __init__(self, file: WsiDicomIO, pixel_data_start: int, frame_count: int):
        super().__init__(file, pixel_data_start, frame_count)
        self._offsets, self._lengths = self._get_tags()

    @property
    def offset_table_type(self):
        return OffsetTableType.TIFF

    def _get_index(self):
        return list(zip(self._offsets, self._lengths))

    def _get_tags(self):
        """Return the tags used for the TIFF table."""
        # Large images will cause `DecompressionBombError` when opened if this is not
        # set to None. Restore it when we are done.
        max_image_pixels_restore = Pillow.MAX_IMAGE_PIXELS
        Pillow.MAX_IMAGE_PIXELS = None
        try:
            image = Pillow.open(self._file)
            tags: Optional[Dict[int, Any]] = getattr(image, "tag_v2", None)
            if tags is None:
                raise EmptyTiffFrameTagsException("File does not contain tiff tags.")
            try:
                offsets: List[int] = tags[TiffTags.TILEOFFSETS.value]
                lengths: List[int] = tags[TiffTags.TILEBYTECOUNTS.value]
            except KeyError as exception:
                raise EmptyTiffFrameTagsException(
                    f"Tiff file is missing required tag {TiffTags(exception.args[0])}."
                ) from exception
            return offsets, lengths
        except UnidentifiedImageError:
            raise EmptyTiffFrameTagsException("File is not a valid tiff file.")
        finally:
            Pillow.MAX_IMAGE_PIXELS = max_image_pixels_restore
